#
#   2025 Fabian Jankowski
#   General helper functions.
#

import logging
import signal
import sys

import matplotlib


def configure_logging(level=logging.INFO):
    """
    Initialize and configure logging.

    Parameters
    ----------
    level: enum
        Requested logging level.
    """

    log = logging.getLogger("fitpdf")
    log.setLevel(logging.DEBUG)
    log.propagate = False

    # ensure a clean root handler
    for h in list(log.handlers):
        log.removeHandler(h)

    console = logging.StreamHandler()
    console.setLevel(level)
    console_formatter = logging.Formatter("%(levelname)s,%(name)s: %(message)s")
    console.setFormatter(console_formatter)
    log.addHandler(console)


def customise_matplotlib_format():
    """
    Customise the matplotlib output formatting.
    """

    matplotlib.rcParams["font.family"] = "serif"
    matplotlib.rcParams["font.size"] = 14.0
    matplotlib.rcParams["lines.markersize"] = 8
    matplotlib.rcParams["legend.frameon"] = False
    # make tickmarks more visible
    matplotlib.rcParams["xtick.major.size"] = 8
    matplotlib.rcParams["xtick.major.width"] = 1.5
    matplotlib.rcParams["xtick.minor.size"] = 4
    matplotlib.rcParams["xtick.minor.width"] = 1.5
    matplotlib.rcParams["ytick.major.size"] = 8
    matplotlib.rcParams["ytick.major.width"] = 1.5
    matplotlib.rcParams["ytick.minor.size"] = 4
    matplotlib.rcParams["ytick.minor.width"] = 1.5


def signal_handler(signum, frame):
    """
    Handle unix signals sent to the program.
    """

    log = logging.getLogger("fitpdf.general_helpers")

    # treat SIGINT/INT/CRTL-C
    if signum == signal.SIGINT:
        log.warning("SIGINT received, stopping the program.")
        sys.exit(1)



#
#   2025 Fabian Jankowski
#   Normal - normal mixture model.
#

import logging

import numpy as np
import pymc as pm

from fitpdf.models.model import Model


class NN(Model):
    name = "Normal - normal"

    def __init__(self):
        """
        Model distribution.
        """

        self.__log = logging.getLogger("fitpdf.models.nn")

        self.ncomp = 2

    def __repr__(self):
        """
        Representation of the object.
        """

        info_dict = {"bla": "XXX"}

        info_str = "{0}".format(info_dict)

        return info_str

    def __str__(self):
        """
        String representation of the object.
        """

        info_str = "{0}: {1}".format(self.name, repr(self))

        return info_str

    def get_model(self, t_data, t_offp, params):
        """
        Construct a mixture model.

        Parameters
        ----------
        t_data: ~np.array of float
            The input data to be fit.
        t_offp: ~np.array of float
            The off-pulse data.
        params: dict
            Additional parameters that influence the processing.

        Returns
        -------
        model: ~pm.Model
            A mixture model.
        """

        data = t_data.copy()
        offp = t_offp.copy()

        # on-pulse mean and std
        onp_mean = np.mean(data)
        onp_std = np.std(data)
        print(f"On-pulse mean: {onp_mean:.5f}")
        print(f"On-pulse std: {onp_std:.5f}")

        # off-pulse mean and std
        offp_mean = np.mean(offp)
        offp_std = np.std(offp)
        print(f"Off-pulse mean: {offp_mean:.5f}")
        print(f"Off-pulse std: {offp_std:.5f}")

        # mixture contributions
        if params["weights"] is None:
            _weights = np.array([0.3, 0.7])
        else:
            _weights = np.array(params["weights"])
            self.__log.info(f"Using custom component weights: {_weights}")

        _weights /= np.sum(_weights)
        self.__log.info(f"Mixture weights: {_weights}")

        coords = {"component": np.arange(2), "obs_id": np.arange(len(data))}

        with pm.Model(coords=coords) as model:
            x = pm.Data("x", data, dims="obs_id")

            # mixture weights
            w = pm.Dirichlet("w", a=_weights, dims="component")

            # priors
            mu = pm.Normal(
                "mu",
                mu=np.array([offp_mean, onp_mean]),
                sigma=np.array([0.1, 1.0]) * onp_std,
                dims="component",
            )
            sigma = pm.HalfNormal(
                "sigma",
                sigma=np.array([offp_std, onp_std]),
                dims="component",
            )

            # 1) normal distribution for nulling
            # 2) normal distribution for pulses
            components = pm.Normal.dist(
                mu=mu,
                sigma=sigma,
                shape=(2,),
            )

            pm.Mixture("obs", w=w, comp_dists=components, observed=x, dims="obs_id")

        return model

    def get_mode(self, posterior, icomp):
        """
        Compute the mode of the model component.

        Returns
        -------
        modes: ~np.array of float
            The mode of the model component.
        """

        mu = posterior["mu"]

        if icomp in [0, 1]:
            mode = mu
        else:
            raise NotImplementedError(f"Component not implemented: {icomp}")

        return mode



#
#   2025 Fabian Jankowski
#   Simulate distributions.
#

import argparse
import logging
import os
import signal
import sys

import numpy as np
import pandas as pd
import pymc as pm

import matplotlib

if "DISPLAY" not in os.environ:
    # set a rendering backend that does not require an X server
    matplotlib.use("Agg")
import matplotlib.pyplot as plt

from fitpdf.general_helpers import (
    configure_logging,
    customise_matplotlib_format,
    signal_handler,
)
from fitpdf.plotting import plot_pedist


def parse_args():
    """
    Parse the commandline arguments.

    Returns
    -------
    args: populated namespace
        The commandline arguments.
    """

    parser = argparse.ArgumentParser(
        description="Simulate distributions.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--nsamp",
        dest="nsamp",
        type=int,
        metavar=("value"),
        default=10000,
        help="Number of random samples to draw from the simulated distribution.",
    )

    parser.add_argument(
        "--randomseed",
        dest="randomseed",
        type=int,
        metavar=("value"),
        default=None,
        help="Enable deterministic mode by providing a seed value for the random number generator. This is useful if you want to fix the underlying distribution when changing the number of samples. The default behaviour is non-deterministic, i.e. the simulation uses different distribution parameters in each run.",
    )

    # options that affect the output formatting
    output = parser.add_argument_group(title="Output formatting")

    output.add_argument(
        "-o",
        "--output",
        dest="output",
        action="store_true",
        default=False,
        help="Output plots to file rather than to screen.",
    )

    args = parser.parse_args()

    return args


def check_args(args):
    """
    Sanity check the commandline arguments.

    Parameters
    ----------
    args: populated namespace
        The commandline arguments.
    """

    log = logging.getLogger("fitpdf.simulate_dist")

    # nsamp
    if not args.nsamp > 100:
        log.error(f"Number of samples is invalid: {args.nsamp}")
        sys.exit(1)

    # randomseed
    if args.randomseed is not None:
        if not args.randomseed > 0:
            log.error(f"Random seed value is invalid: {args.randomseed}")
            sys.exit(1)


#
# MAIN
#


def main():
    # start signal handler
    signal.signal(signal.SIGINT, signal_handler)

    # set up logging
    configure_logging()
    log = logging.getLogger("fitpdf.simulate_dist")

    # handle command line arguments
    args = parse_args()

    # sanity check command line arguments
    check_args(args)

    # tweak the matplotlib output formatting
    customise_matplotlib_format()

    params = {
        "dpi": 300,
        "nbin": "auto",
        "label": "on",
        "nsamp": args.nsamp,
        "output": args.output,
        "title": None,
    }

    # switch between non-deterministic (default) and deterministic mode
    if args.randomseed is None:
        rng = None
    else:
        rng = np.random.default_rng(args.randomseed)

    weights = [0.2, 0.3, 0.5]
    weights /= np.sum(weights)
    mu = [0.0, 0.5, 0.0]
    sigma = [1.0, 2.0, 0.5]

    print(f"Mixture weights: {weights}")

    # off-pulse
    foff = pm.Normal.dist(mu=mu[0], sigma=sigma[0])

    foff_samples = pm.draw(foff, draws=params["nsamp"], random_seed=rng)

    # on-pulse
    fon = pm.Mixture.dist(
        w=weights,
        comp_dists=[
            pm.Normal.dist(mu=mu[0], sigma=sigma[0]),
            pm.Normal.dist(mu=mu[1], sigma=sigma[1]),
            pm.LogNormal.dist(mu=mu[2], sigma=sigma[2]),
        ],
    )

    fon_samples = pm.draw(fon, draws=params["nsamp"], random_seed=rng)

    # write to disk
    _temp = {
        "rotation": np.arange(params["nsamp"]),
        "zapped": np.zeros(params["nsamp"]).astype(int),
        "fluence_on": fon_samples,
        "nbin_on": 64,
        "fluence_off": foff_samples,
        "nbin_off": 64,
        "fluence_off_same": foff_samples,
        "nbin_off_same": 64,
    }
    _df = pd.DataFrame(_temp)
    _df.to_csv("simulated_fluences.csv", index=False)

    params["outfile"] = "simulated_pdf.pdf"
    plot_pedist(fon_samples, foff_samples, params)

    plt.show()

    log.info("All done.")


if __name__ == "__main__":
    main()



#
#   2025 Fabian Jankowski
#   Compare fits.
#

import argparse
import logging
import os
import signal
import sys

import arviz as az

if "DISPLAY" not in os.environ:
    # set a rendering backend that does not require an X server
    matplotlib.use("Agg")
import matplotlib.pyplot as plt

from fitpdf.general_helpers import (
    configure_logging,
    signal_handler,
)
from fitpdf.plotting import plot_fit_comparison


def parse_args():
    """
    Parse the commandline arguments.

    Returns
    -------
    args: populated namespace
        The commandline arguments.
    """

    parser = argparse.ArgumentParser(
        description="Compare fits.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "files",
        type=str,
        nargs="+",
        help="Names of files to process. The input files must be InferenceData files produced by fitpdf-fit.",
    )

    # options that affect the output formatting
    output = parser.add_argument_group(title="Output formatting")

    output.add_argument(
        "-o",
        "--output",
        dest="output",
        action="store_true",
        default=False,
        help="Output plots to file rather than to screen.",
    )

    args = parser.parse_args()

    return args


def check_args(args):
    """
    Sanity check the commandline arguments.

    Parameters
    ----------
    args: populated namespace
        The commandline arguments.
    """

    log = logging.getLogger("fitpdf.compare_fits")

    # check that files exist
    for item in args.files:
        if not os.path.isfile(item):
            log.error(f"File does not exist: {item}")
            sys.exit(1)


#
# MAIN
#


def main():
    # start signal handler
    signal.signal(signal.SIGINT, signal_handler)

    # set up logging
    configure_logging()
    log = logging.getLogger("fitpdf.compare_fits")

    # handle command line arguments
    args = parse_args()

    # sanity check command line arguments
    check_args(args)

    params = {
        "dpi": 300,
        "output": args.output,
    }

    idatas = {}

    for item in args.files:
        print(f"Processing: {item}")
        _idata = az.from_netcdf(item)
        _label = item.rstrip(".nc")
        _label = _label.lstrip("idata_")
        idatas[_label] = _idata

    df_comp = az.compare(idatas)
    print(df_comp)

    plot_fit_comparison(df_comp, params)

    plt.show()

    log.info("All done.")


if __name__ == "__main__":
    main()



#
#   2025 Fabian Jankowski
#   Distribution models.
#

from fitpdf.models.ll import LL
from fitpdf.models.model import Model
from fitpdf.models.nl import NL
from fitpdf.models.nn import NN
from fitpdf.models.nnl import NNL
from fitpdf.models.nnp import NNP



#
#   2025 Fabian Jankowski
#   Plotting related helper functions.
#

import logging

import arviz as az
import corner
from KDEpy import FFTKDE, TreeKDE
from KDEpy.bw_selection import improved_sheather_jones
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
import numpy as np
import pymc as pm
from scipy import stats

from fitpdf.stats import get_adaptive_bandwidth


def plot_chains(idata, params):
    """
    Plot the chains.

    Parameters
    ----------
    idata: ~az.InterferenceData
        The input data.
    params: dict
        Additional parameters that influence the processing.
    """

    az.plot_trace(idata, trace_kwargs={"rasterized": True})

    fig = plt.gcf()
    fig.tight_layout()

    # output plot to file
    if params["output"]:
        fig.savefig(
            "chains.pdf",
            bbox_inches="tight",
            dpi=params["dpi"],
        )

        plt.close(fig)


def plot_corner(idata, params):
    """
    Make a corner plot.

    Parameters
    ----------
    idata: ~az.InterferenceData
        The input data.
    params: dict
        Additional parameters that influence the processing.
    """

    # defaults
    bins = 40
    fontsize_before = matplotlib.rcParams["font.size"]
    hist_kwargs = None
    labelpad = 0.125
    max_n_ticks = 5
    plot_datapoints = False
    show_titles = True
    smooth = False

    if params["publish"]:
        hist_kwargs = {"lw": 2.0}
        labelpad = 0.475
        max_n_ticks = 2
        matplotlib.rcParams["font.size"] = 34.0
        show_titles = False
        smooth = True

    fig = corner.corner(
        idata,
        bins=bins,
        hist_kwargs=hist_kwargs,
        labelpad=labelpad,
        max_n_ticks=max_n_ticks,
        plot_datapoints=plot_datapoints,
        quantiles=[0.16, 0.5, 0.84],
        show_titles=show_titles,
        smooth=smooth,
        title_kwargs={"fontsize": 10},
    )

    # output plot to file
    if params["output"]:
        fig.savefig(
            "corner.pdf",
            bbox_inches="tight",
            dpi=params["dpi"],
        )

        plt.close(fig)

    # reset
    matplotlib.rcParams["font.size"] = fontsize_before


def plot_adaptive_bandwidths(t_data, t_bandwidths, params):
    """
    Visualise the KDE adaptive bandwidths.

    Parameters
    ----------
    t_data: ~np.array of float
        The input data of the KDE.
    t_bandwidths: ~np.array of float
        The adaptive KDE bandwidths for each data point.
    params: dict
        Other parameters that influence the processing.
    """

    data = t_data.copy()
    bandwidths = t_bandwidths.copy()

    ys = np.arange(len(data)) / len(data)

    fig = plt.figure()
    ax = fig.add_subplot()

    ax.scatter(data, ys, marker="+", color="black", label="data", zorder=5)

    ax.errorbar(
        x=data,
        y=ys,
        xerr=bandwidths,
        color="C1",
        elinewidth=0.5,
        label="bandwidth",
        linestyle="none",
        marker=".",
        zorder=4,
    )

    ax.legend(loc="best", frameon=False)
    ax.set_xlabel("Value")

    fig.tight_layout()

    # output plot to file
    if params["output"]:
        fig.savefig(
            "kde_adaptive_bandwidths.pdf",
            bbox_inches="tight",
            dpi=params["dpi"],
        )

        plt.close(fig)


def plot_pedist(fon, foff, params):
    """
    Plot the pulse-energy distribution data samples.

    Parameters
    ----------
    fon: ~np.array of float
        The on-pulse fluence samples.
    foff: ~np.array of float
        The off-pulse fluence samples.
    params: dict
        Other parameters that influence the processing.
    """

    fig = plt.figure()
    ax = fig.add_subplot()

    if params["label"] is None:
        _label = "on"
    else:
        _label = params["label"]

    ax.hist(
        fon,
        bins=params["nbin"],
        color="black",
        density=True,
        histtype="step",
        label=_label,
        lw=2,
        zorder=5,
    )

    ax.hist(
        foff,
        bins=params["nbin"],
        color="dimgrey",
        density=True,
        histtype="stepfilled",
        label="off",
        zorder=3,
        alpha=0.3,
    )

    ax.legend(loc="best", frameon=False)
    if params["title"] is not None:
        ax.set_title(params["title"])
    ax.set_xlabel(r"$F \: / \: \left< F_\mathrm{on} \right>$")
    ax.set_ylabel("PDF")
    ax.set_yscale("log")

    fig.tight_layout()

    # output plot to file
    if params["output"]:
        fig.savefig(
            params["outfile"],
            bbox_inches="tight",
            dpi=params["dpi"],
        )

        plt.close(fig)


def plot_fit(mobj, model, idata, offp, params):
    """
    Plot the distribution fit.

    Parameters
    ----------
    mobj: ~fitpdf.models.Model
        Model object.
    idata: ~az.InterferenceData
        The input data.
    offp: ~pd.DataFrame
        The off-pulse data.
    params: dict
        Additional parameters that influence the processing.
    """

    log = logging.getLogger("fitpdf.plotting")

    obs_data = np.sort(idata.observed_data["obs"].values)

    fig = plt.figure()
    ax = fig.add_subplot()

    # plot the observed data
    # kernel density estimate using adaptive bandwidth
    isj_bw_data = improved_sheather_jones(obs_data.reshape(obs_data.shape[0], -1))
    log.info(f"ISJ kernel bandwidth data: {isj_bw_data:.5f}")

    min_bw_data = 5.0 * isj_bw_data
    log.info(f"Minimum bandwidth data: {min_bw_data:.5f}")

    bandwidths = get_adaptive_bandwidth(obs_data, min_bw=min_bw_data)
    log.info(f"Bandwidths: {bandwidths}")
    plot_adaptive_bandwidths(obs_data, bandwidths, params)

    kde_x_data, kde_y_data = (
        TreeKDE(kernel="gaussian", bw=bandwidths).fit(obs_data).evaluate()
    )

    if params["label"] is None:
        _label = "on"
    else:
        _label = params["label"]

    ax.plot(kde_x_data, kde_y_data, color="black", lw=2, label=_label, zorder=4)

    # rug plot
    # use data coordinates in horizontal and axis coordinates in vertical direction
    trans = transforms.blended_transform_factory(ax.transData, ax.transAxes)
    ax.scatter(
        obs_data,
        [0.99 for _ in range(len(obs_data))],
        marker="|",
        color="black",
        lw=0.3,
        transform=trans,
        alpha=0.1,
        rasterized=True,
    )

    # off pulse
    _isj_bw = improved_sheather_jones(offp.reshape(offp.shape[0], -1))
    _bandwidths = get_adaptive_bandwidth(offp, min_bw=5.0 * _isj_bw)
    kde_x, kde_y = TreeKDE(kernel="gaussian", bw=_bandwidths).fit(offp).evaluate()

    ax.fill_between(
        x=kde_x,
        y1=kde_y,
        y2=0,
        facecolor="dimgrey",
        edgecolor="none",
        label="off",
        lw=0,
        alpha=0.2,
        zorder=3,
    )

    # plot the best-fitting model
    # use fftkde here with non-adaptive bandwidth for speed
    samples = idata.posterior_predictive["obs"].values.reshape(-1)
    _mask = (samples >= kde_x_data.min()) & (samples <= kde_x_data.max())
    kde_y = (
        FFTKDE(kernel="gaussian", bw=min_bw_data)
        .fit(samples[_mask])
        .evaluate(kde_x_data)
    )

    ax.plot(kde_x_data, kde_y, color="firebrick", lw=1.5, label="model", zorder=5)

    # perform a two-sample kolmogorov-smirnov test
    ks_test = stats.ks_2samp(obs_data, samples, axis=None)
    print("\nTwo-sample KS test data vs model:")
    print(f"Statistic: {ks_test.statistic:.3f}")
    print(f"p-value: {ks_test.pvalue:.3f}")
    print(f"Location: {ks_test.statistic_location:.3f}")
    print(f"Sign: {ks_test.statistic_sign}")

    # plot the individual pp draws
    _ndraw = 50
    rng = np.random.default_rng()
    idxs_chain = rng.integers(
        low=0, high=len(idata.posterior_predictive["chain"]), size=_ndraw
    )
    idxs_draw = rng.integers(
        low=0, high=len(idata.posterior_predictive["draw"]), size=_ndraw
    )

    for ichain, idraw in zip(idxs_chain, idxs_draw):
        samples = (
            idata.posterior_predictive["obs"]
            .isel(chain=ichain, draw=idraw)
            .values.reshape(-1)
        )
        _mask = (samples >= kde_x_data.min()) & (samples <= kde_x_data.max())
        _bandwidths = get_adaptive_bandwidth(samples[_mask], min_bw=min_bw_data)
        kde_y = (
            TreeKDE(kernel="gaussian", bw=_bandwidths)
            .fit(samples[_mask])
            .evaluate(kde_x_data)
        )

        ax.plot(
            kde_x_data,
            kde_y,
            color="firebrick",
            lw=0.5,
            zorder=3.5,
            alpha=0.1,
            rasterized=True,
        )

    # plot the individual model components
    assert hasattr(mobj, "ncomp")

    for icomp in range(mobj.ncomp):
        # switch off the other component distributions
        # dirichlet distribution must be normalised to unity
        _weights = np.zeros(mobj.ncomp)
        _weights[icomp] = 1.0

        fixed_params = {"w": _weights}
        with pm.do(model, fixed_params):
            thinned_idata = idata.sel(draw=slice(None, None, 20))
            pp = pm.sample_posterior_predictive(thinned_idata, var_names=["obs"])

            samples = pp.posterior_predictive["obs"].values.reshape(-1)
            _mask = (samples >= kde_x_data.min()) & (samples <= kde_x_data.max())
            kde_y = (
                FFTKDE(kernel="gaussian", bw=min_bw_data)
                .fit(samples[_mask])
                .evaluate(kde_x_data)
            )

            # scale kde_y to true component weight
            _true_w = (
                idata.posterior["w"]
                .sel(component=icomp)
                .mean(dim=("chain", "draw"))
                .values
            )
            true_kde_y = _true_w * kde_y

            ax.plot(
                kde_x_data,
                true_kde_y,
                label=f"c{icomp}",
                lw=1,
                zorder=6,
            )

    ax.legend(loc="best", frameon=False)
    if params["title"] is not None:
        ax.set_title(params["title"])
    ax.set_xlabel(r"$F \: / \: \left< F_\mathrm{on} \right>$")
    ax.set_ylabel("PDF")
    ax.set_yscale("log")

    ax.set_xlim(left=1.25 * obs_data.min(), right=1.05 * obs_data.max())

    # set the limits to a bin count of unity
    _density_data, _ = np.histogram(
        obs_data,
        bins=params["nbin"],
        density=True,
    )

    _mask = np.isfinite(_density_data) & (_density_data > 0)
    min_density = np.min(_density_data[_mask])
    max_density = np.max(_density_data[_mask])

    ax.set_ylim(bottom=0.7 * min_density, top=2.0 * max_density)

    fig.tight_layout()

    # output plot to file
    if params["output"]:
        fig.savefig(
            "pedist_fit.pdf",
            bbox_inches="tight",
            dpi=params["dpi"],
        )

        plt.close(fig)


def plot_prior_predictive(idata, t_data, t_offp, params):
    """
    Plot the prior predictive samples.

    Parameters
    ----------
    idata: ~az.InterferenceData
        The input data.
    offp: ~pd.DataFrame
        The off-pulse data.
    params: dict
        Additional parameters that influence the processing.
    """

    data = t_data.copy()
    offp = t_offp.copy()

    # 1) density plot for all parameters
    az.plot_density(
        idata,
        group="prior",
        shade=0.1,
    )

    fig = plt.gcf()
    fig.tight_layout()

    # output plot to file
    if params["output"]:
        fig.savefig(
            "prior_predictive_densities.pdf",
            bbox_inches="tight",
            dpi=params["dpi"],
        )

        plt.close(fig)

    # 2) comparison with the data
    fig = plt.figure()
    ax = fig.add_subplot()

    bins = np.linspace(data.min(), data.max(), num=params["nbin"])

    # data
    if params["label"] is None:
        label = "on"
    else:
        label = params["label"]

    ax.hist(
        data,
        bins=bins,
        color="black",
        density=True,
        histtype="step",
        label=label,
        lw=2,
        zorder=4,
    )

    # off-pulse
    ax.hist(
        offp,
        bins=params["nbin"],
        color="dimgrey",
        density=True,
        histtype="stepfilled",
        linewidth=2,
        label="off",
        zorder=3,
        alpha=0.4,
    )

    # prior predictive samples
    ax.hist(
        idata.prior_predictive["obs"].values.reshape(-1),
        bins=bins,
        color="C1",
        density=True,
        histtype="step",
        label="prior predictive",
        lw=2,
        zorder=5,
    )

    ax.legend(loc="best", frameon=False)
    ax.set_xlabel(r"$F \: / \: \left< F_\mathrm{on} \right>$")
    ax.set_ylabel("PDF")
    ax.set_yscale("log")

    fig.tight_layout()

    # output plot to file
    if params["output"]:
        fig.savefig(
            "prior_predictive_data.pdf",
            bbox_inches="tight",
            dpi=params["dpi"],
        )

        plt.close(fig)


def plot_fit_comparison(df_comp, params):
    """
    Plot a fit comparison.

    Parameters
    ----------
    df_comp: ~pd.DataFrame
        The fit comparison data.
    params: dict
        Additional parameters that influence the processing.
    """

    figsize = (7.4, len(df_comp.index))
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot()

    az.plot_compare(df_comp, ax=ax, insample_dev=True, legend=True, plot_ic_diff=True)

    fig.tight_layout()

    # output plot to file
    if params["output"]:
        fig.savefig(
            "fit_comparison.pdf",
            bbox_inches="tight",
            dpi=params["dpi"],
        )

        plt.close(fig)



#
#   2025 Fabian Jankowski
#   Fit measured distribution data.
#

import argparse
import logging
import os
import signal
import sys

import arviz as az

# switch between interactive and non-interactive mode
import matplotlib

if "DISPLAY" not in os.environ:
    # set a rendering backend that does not require an X server
    matplotlib.use("Agg")
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pymc as pm

from fitpdf.general_helpers import (
    configure_logging,
    customise_matplotlib_format,
    signal_handler,
)
import fitpdf.models as fmodels
from fitpdf.plotting import (
    plot_chains,
    plot_corner,
    plot_fit,
    plot_pedist,
    plot_prior_predictive,
)


def parse_args():
    """
    Parse the commandline arguments.

    Returns
    -------
    args: populated namespace
        The commandline arguments.
    """

    parser = argparse.ArgumentParser(
        description="Fit distribution data.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "filename",
        type=str,
        help="Name of file to process. The input file must be produced by the fluence time series option of plot-profilestack.",
    )

    parser.add_argument(
        "--mean",
        dest="mean",
        type=float,
        metavar=("value"),
        default=None,
        help="The global mean fluence by which to divide the histograms. The default behaviour is to determine it automatically from the on-pulse fluence data.",
    )

    # fit parameter options
    fitp = parser.add_argument_group(title="Fit parameters")

    fitp.add_argument(
        "--fast",
        dest="fast",
        action="store_true",
        default=False,
        help="Enable fast processing. This reduces the number of MCMC samples drastically and is recommended against for publication-quality fits.",
    )

    fitp.add_argument(
        "--model",
        dest="model",
        choices=["NL", "NN", "NNL", "NNP"],
        default="NNL",
        help="Use the specified distribution model, where N denotes a Normal, L a Lognormal, and P a powerlaw (Pareto) component. For instance, the default NNL model consists of two Normal and one Lognormal distributions.",
    )

    fitp.add_argument(
        "--weights",
        dest="weights",
        type=float,
        metavar=("value"),
        nargs="+",
        default=None,
        help="Override the default component distribution weights in the model prior. This is sometimes useful to ensure convergence of the fit. The weights are given as simple floating point numbers (not percentages) and must sum to unity. For instance, [0.2, 0.3, 0.5] assigns an average prior weight of 20, 30, and 50 per cent to each of the component distributions, respectively. The number of weights specified must match the number of model components, e.g. three for the NNL model.",
    )

    # options that affect the output formatting
    output = parser.add_argument_group(title="Output formatting")

    output.add_argument(
        "--label",
        dest="label",
        type=str,
        metavar=("name"),
        default=None,
        help="The label to use for the input file.",
    )

    output.add_argument(
        "--nbin",
        dest="nbin",
        type=int,
        metavar=("value"),
        default=50,
        help="The number of histogram bins to use.",
    )

    output.add_argument(
        "-o",
        "--output",
        dest="output",
        action="store_true",
        default=False,
        help="Output plots to file rather than to screen.",
    )

    output.add_argument(
        "--title",
        dest="title",
        type=str,
        metavar=("text"),
        default=None,
        help="Set a custom figure title.",
    )

    args = parser.parse_args()

    return args


def check_args(args):
    """
    Sanity check the commandline arguments.

    Parameters
    ----------
    args: populated namespace
        The commandline arguments.
    """

    log = logging.getLogger("fitpdf.fit_pdf")

    # check the mean
    if args.mean is not None:
        if args.mean > 0:
            pass
        else:
            log.error(f"The mean fluence is invalid: {args.mean}")
            sys.exit(1)

    # nbin
    if not args.nbin > 1:
        log.error(f"The number of bins is invalid: {args.nbin}")
        sys.exit(1)

    # weights
    if args.weights is not None:
        if len(args.weights) == len(args.model) and sum(args.weights) == 1.0:
            pass
        else:
            log.error(f"The weights are invalid: {args.weights}")
            sys.exit(1)

    # check that file exist
    if not os.path.isfile(args.filename):
        log.error(f"File does not exist: {args.filename}")
        sys.exit(1)


def get_clean_data(t_df):
    """
    Clean the data.

    Returns
    -------
    fon: ~np.array of float
        The cleaned on-pulse fluence data.
    foff: ~np.array of float
        The cleaned off-pulse fluence data.
    """

    log = logging.getLogger("fitpdf.fit_pdf")

    df = t_df.copy()

    # exclude rfi zapped data
    mask_zapped = df["zapped"].astype(bool)
    mask_good = np.logical_not(mask_zapped)

    # on-pulse
    fon = df["fluence_on"].to_numpy()
    mask = mask_good & np.isfinite(fon)

    log.info(f"Size of Fon before cleaning: {fon.size}")
    fon = fon[mask]
    fon = np.sort(fon)
    log.info(f"Size of Fon after cleaning: {fon.size}")

    # off-pulse
    if df["nbin_off_same"].iat[0] == df["nbin_on"].iat[0]:
        foff = df["fluence_off_same"].to_numpy()
    else:
        foff = (df["fluence_off_same"] * df["nbin_on"] / df["nbin_off_same"]).to_numpy()
        log.warning(
            "Off-pulse phase range is too small: {0}, {1}. Scaling the off-pulse fluences.".format(
                df["nbin_off_same"].iat[0], df["nbin_on"].iat[0]
            )
        )

    log.info(f"Size of Foff before cleaning: {foff.size}")
    foff = foff[mask]
    foff = np.sort(foff)
    log.info(f"Size of Foff after cleaning: {foff.size}")

    assert np.all(np.isfinite(fon))
    assert np.all(np.isfinite(foff))
    assert fon.size == foff.size

    return fon, foff


def fit_pedist(t_data, t_offp, params):
    """
    Fit pulse-energy distribution data.

    Parameters
    ----------
    t_data: ~np.array of float
        The on-pulse data.
    t_offp: ~np.array of float
        The off-pulse data.
    params: dict
        Additional parameters that influence the processing.
    """

    log = logging.getLogger("fitpdf.fit_pdf")

    data = t_data.copy()
    offp = t_offp.copy()

    # model selection
    if params["model"] == "NL":
        mobj = fmodels.NL()
    elif params["model"] == "NN":
        mobj = fmodels.NN()
    elif params["model"] == "NNL":
        mobj = fmodels.NNL()
    elif params["model"] == "NNP":
        log.error("The NNP model is currently disabled.")
        sys.exit(1)
    else:
        raise NotImplementedError("Model not implemented: %s", params["model"])

    model = mobj.get_model(data, offp, params)

    print(f"All RVs: {model.basic_RVs}")
    print(f"Free RVs: {model.free_RVs}")
    print(f"Observed RVs: {model.observed_RVs}")
    print(f"Initial point: {model.initial_point()}")

    config = {}

    if params["fast"]:
        config["draws"] = 700
        config["tune"] = 700
    else:
        config["draws"] = 10000
        config["tune"] = 2000

    with model:
        idata = pm.sample(
            draws=config["draws"],
            tune=config["tune"],
            chains=4,
            init="advi+adapt_diag",
            nuts={"target_accept": 0.9},
        )
        pm.compute_log_likelihood(idata)

    _df_result = az.summary(idata)
    print(_df_result)
    _df_result.to_csv("fit_result.csv")

    plot_chains(idata, params)
    plot_corner(idata, params)

    # compute prior predictive samples
    with model:
        # sample all the parameters
        pp = pm.sample_prior_predictive()

    assert hasattr(pp, "prior_predictive")

    plot_prior_predictive(pp, data, offp, params)

    # compute posterior predictive samples
    thinned_idata = idata.sel(draw=slice(None, None, 20))

    with model:
        pp = pm.sample_posterior_predictive(thinned_idata, var_names=["obs"])
        idata.extend(pp)

    assert hasattr(idata, "posterior_predictive")

    # save idata to file
    _filename = "idata_{0}.nc".format(params["model"])
    az.to_netcdf(idata, _filename)

    plot_fit(mobj, model, idata, offp, params)

    # output the fit parameters
    print("\nFit parameters")
    for icomp in range(mobj.ncomp):
        # weight
        _samples = idata.posterior["w"]
        quantiles = _samples.sel(component=icomp).quantile(
            q=[0.16, 0.5, 0.84], dim=("chain", "draw")
        )
        error = np.maximum(
            np.abs(quantiles[1] - quantiles[0]), np.abs(quantiles[2] - quantiles[1])
        )
        weight = {"value": quantiles[1], "error": error}

        # mode
        _samples = mobj.get_mode(idata.posterior, icomp)
        quantiles = _samples.sel(component=icomp).quantile(
            q=[0.16, 0.5, 0.84], dim=("chain", "draw")
        )
        error = np.maximum(
            np.abs(quantiles[1] - quantiles[0]), np.abs(quantiles[2] - quantiles[1])
        )
        mode = {"value": quantiles[1], "error": error}

        print(
            "Component {0}: {1:.3f} +- {2:.3f} %, {3:.3f} +- {4:.3f}".format(
                icomp,
                100.0 * weight["value"],
                100.0 * weight["error"],
                mode["value"],
                mode["error"],
            )
        )


#
# MAIN
#


def main():
    # start signal handler
    signal.signal(signal.SIGINT, signal_handler)

    # set up logging
    configure_logging()
    log = logging.getLogger("fitpdf.fit_pdf")

    # handle command line arguments
    args = parse_args()

    # sanity check command line arguments
    check_args(args)

    # tweak the matplotlib output formatting
    customise_matplotlib_format()

    params = {
        "dpi": 300,
        "fast": args.fast,
        "label": args.label,
        "mean": args.mean,
        "model": args.model,
        "nbin": args.nbin,
        "output": args.output,
        "publish": False,
        "title": args.title,
        "weights": args.weights,
    }

    print(f"Processing: {args.filename}")
    df = pd.read_csv(args.filename)
    df["filename"] = args.filename

    # sanity check
    _fields = [
        "rotation",
        "zapped",
        "fluence_on",
        "nbin_on",
        "fluence_off",
        "nbin_off",
        "fluence_off_same",
        "nbin_off_same",
    ]
    assert all(item in df.columns for item in _fields)

    _fon, _foff = get_clean_data(df)

    for item, _label in zip([_fon, _foff], ["On", "Off"]):
        print(
            "{0}-pulse mean, median, std, number of samples: {1:.5f}, {2:.5f}, {3:.5f}, {4}".format(
                _label, np.mean(item), np.median(item), np.std(item), len(item)
            )
        )

    # standardise the data by dividing by the mean on-pulse fluence
    # we should also divide by the on-pulse standard deviation for a
    # proper data standardisation (unity mean and unity standard deviation)
    # however, then we must keep track of the parameters and transform the inference data back later
    # we express the priors in data units (on and off-pulse mean fluence) instead here
    # this avoids the back transformation later
    if args.mean is None:
        _global_mean = np.mean(_fon)
    else:
        _global_mean = params["mean"]

    print(f"Global mean on-pulse fluence: {_global_mean}")

    params["outfile"] = "pedist_pdf.pdf"
    plot_pedist(_fon / _global_mean, _foff / _global_mean, params)

    fit_pedist(_fon / _global_mean, _foff / _global_mean, params)

    plt.show()

    log.info("All done.")


if __name__ == "__main__":
    main()
